/* vim:fdm=marker ts=4 et ai
 * {{{
 *
 * (c) by Alexander Neumann <alexander@bumpern.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 }}} */

#include "7segmentclock.h"
#include "syslog.h"

#ifdef DEBUG_7SEG
#include "uart.h"
#endif

uint8_t l7seg_index;

/* global variables */
char *l7seg_buffer; //this will be displayed instead of time
uint32_t l7dontdisplaytime; //until this timestamp the 7segment clock will show the buffer contents instead of time

void l7seg_init(void) {
	l7seg_index = 5; //next index will be 0
	l7dontdisplaytime = 0;
	l7seg_buffer=malloc(6);
	//the used bits of port D will be output
	L7SEG_DDR |= ((1 << L7SEG_PORT_SEG_CLEAR) | (1 << L7SEG_PORT_SEG_CLOCK) | (1 << L7SEG_PORT_SEG_DATA) | (1 << L7SEG_PORT_IDX_CLEAR) | (1 << L7SEG_PORT_IDX_CLOCK) | (1 << L7SEG_PORT_IDX_ENABLE));
	//set the reset bits of 74HC163 and 74HC164, the output enable bit of 74HC138 and the clock input of 74HC163 to high
	L7SEG_PORT |= ((1 << L7SEG_PORT_SEG_CLEAR) | (1 << L7SEG_PORT_IDX_CLEAR) | (1 << L7SEG_PORT_IDX_ENABLE) | (1 << L7SEG_PORT_SEG_CLOCK));
	//set the data input of 74HC164 (segment data) and the clock input of 74HC163 (display index) to low
	L7SEG_PORT &= ~((1 << L7SEG_PORT_SEG_DATA) | (1 << L7SEG_PORT_IDX_CLOCK));
}

void l7seg_periodic(void)
/* {{{ */ {
	uint8_t segdata=0;


#ifdef DEBUG_7SEG
   	uart_puts_P("7seg: updating display\r\n");
#endif

    /* convert time to something useful */
    struct clock_datetime_t d;
    clock_current_datetime(&d);

	//Disable all output while changing it (set enable bit of 74HC138 to low)
	L7SEG_PORT &= ~(1 << L7SEG_PORT_IDX_ENABLE);

	//increment output position
	l7seg_index++;
	if (l7seg_index == 6) {
		//we reached the position after the last display index -> reset to first display index
		l7seg_index = 0;
		//set reset input of 74HC164 to low
		L7SEG_PORT &= ~(1 << L7SEG_PORT_IDX_CLEAR);
		//send a short low pulse to the clock input of the index counter
		//low-to-high edge triggers reset
		L7SEG_PORT &= ~(1 << L7SEG_PORT_IDX_CLOCK);
		L7SEG_DELAY();
		L7SEG_PORT |= (1 << L7SEG_PORT_IDX_CLOCK);
		L7SEG_DELAY();
		//set reset input of 74HC164 to high again
		L7SEG_PORT |= (1 << L7SEG_PORT_IDX_CLEAR);
	} else {
		//send a short low pulse to the clock input of the index counter
		L7SEG_PORT &= ~(1 << L7SEG_PORT_IDX_CLOCK);
		L7SEG_DELAY();
		L7SEG_PORT |= (1 << L7SEG_PORT_IDX_CLOCK);
	}

	if (clock_global.current_time > l7dontdisplaytime) {

		switch (l7seg_index) {
			case 0:
				segdata = l7seg_segvalue(d.hour/10);
				break;
			case 1:
				segdata = l7seg_segvalue(d.hour-10*(d.hour/10)) | 0x01;
				break;
			case 2:
				segdata = l7seg_segvalue(d.min/10);
				break;
			case 3:
				segdata = l7seg_segvalue(d.min-10*(d.min/10)) | 0x01;
				break;
			case 4:
				segdata = l7seg_segvalue(d.sec/10);
				break;
			case 5:
				segdata = l7seg_segvalue(d.sec-10*(d.sec/10));
				break;
		}
	
	} else {

		segdata = l7seg_segvalue(l7seg_buffer[l7seg_index]);

	}

	//clear 7segment-register (send a short low pulse to the clear-port of the segment register)
	L7SEG_PORT &= ~(1 << L7SEG_PORT_SEG_CLEAR);
	L7SEG_DELAY();
	L7SEG_PORT |= (1 << L7SEG_PORT_SEG_CLEAR);

	//shift 7segment data bitwise into register
	for (uint8_t i=0; i<=7; i++) {
		uint8_t bit=segdata & 1;
		segdata = (segdata >> 1);
		//set the bit value to the data input of the segment register
		if (bit == 1) {
			L7SEG_PORT |= (1 << L7SEG_PORT_SEG_DATA);
		} else {
			L7SEG_PORT &= ~(1 << L7SEG_PORT_SEG_DATA);
		}
		//when data is valid, send a short high pulse to the clock input of the segment register
		//(clock input of 74HC164 is triggered with positive going transition)
		L7SEG_PORT |= (1 << L7SEG_PORT_SEG_CLOCK);
		L7SEG_DELAY();
		L7SEG_PORT &= ~(1 << L7SEG_PORT_SEG_CLOCK);
	}

	//Enable output again (set enable bit to high again)
	L7SEG_PORT |= (1 << L7SEG_PORT_IDX_ENABLE);

} /* }}} */

uint8_t l7seg_segvalue(uint8_t decval)
{
    uint8_t i=0;

	switch (decval) {
		case 0:
			i=0xfc;
			break;
		case 1:
			i=0x60;
			break;
		case 2:
			i=0xda;
			break;
		case 3:
			i=0xf2;
			break;
		case 4:
			i=0x66;
			break;
		case 5:
			i=0xb6;
			break;
		case 6:
			i=0xbe;
			break;
		case 7:
			i=0xe0;
			break;
		case 8:
			i=0xfe;
			break;
		case 9:
			i=0xf6;
			break;
		case 'a':
			i=0xee;
			break;
		case 'e':
			i=0x9e;
			break;
		case 'i':
			i=0x60;
			break;
		case 'o':
			i=0xfc;
			break;
		case 'p':
			i=0xce;
			break;
		case 'r':
			i=0x0a;
			break;
		case 's':
			i=0xb6;
			break;
		case 't':
			i=0x1e;
			break;
	}
	return i;
}
