/*
 * Copyright (c) 2007, Stefan Helbing.
 * Based on telnetd from uIP stack 1.0 by:
 * Copyright (c) 2003, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: telnetd.c,v 1.2 2006/06/07 09:43:54 adam Exp $
 *
 */

#include <avr/eeprom.h>
#include <string.h>
#include <util/delay.h>

#include "uip/uip.h"
#include "telnetd.h"
#include "syslog.h"
#include "eeprom.h"

#ifdef DEBUG_TELNETD
#include "uart.h"
#endif

#define ISO_nl       0x0a
#define ISO_cr       0x0d

/* global variables */
#ifdef USE_7SEGMENT
char *l7seg_buffer; //this will be displayed instead of time
uint32_t l7dontdisplaytime; //until this timestamp the 7segment clock will show the buffer contents instead of time
#endif

char telnetd_name[8][9];

struct telnetd_line {
	char line[TELNETD_CONF_LINELEN];
};

struct ptentry {
	char *commandstr;
	void (* pfunc)(struct telnetd_connection_state_t *state);
};

/* local prototypes */
unsigned short send_str_P(void *data);
void shell_prompt(struct telnetd_connection_state_t *state, const char *str);
void shell_output(struct telnetd_connection_state_t *state, const char *str1, const char *str2);
void shell_input(struct telnetd_connection_state_t *state);
static void telnetd_io(struct telnetd_connection_state_t *state);
static void telnetd_rst(struct telnetd_connection_state_t *state);
static void help(struct telnetd_connection_state_t *state);
static void unknown(struct telnetd_connection_state_t *state);
void shell_quit(struct telnetd_connection_state_t *state);
static void sendline(struct telnetd_connection_state_t *state, char *line);

static struct ptentry parsetab[] = {
	{"io", telnetd_io},
	{"rst", telnetd_rst},
	{"help", help},
	{"exit", shell_quit},
	{"?", help},
	{NULL, unknown} 	/* Default action */
};


MEMB(linemem, struct telnetd_line, TELNETD_CONF_NUMLINES);

#define STATE_NORMAL 0
#define STATE_IAC    1
#define STATE_WILL   2
#define STATE_WONT   3
#define STATE_DO     4
#define STATE_DONT   5
#define STATE_CLOSE  6

//static struct telnetd_state s;

#define TELNET_NOP	 241
#define TELNET_WILL  251
#define TELNET_WONT  252
#define TELNET_DO    253
#define TELNET_DONT  254
#define TELNET_IAC   255

/*---------------------------------------------------------------------------*/
void memb_init(struct memb_blocks *m) {
	memset(m->count, 0, m->num);
	memset(m->mem, 0, m->size * m->num);
}
/*---------------------------------------------------------------------------*/
void * memb_alloc(struct memb_blocks *m) {
	int i;

	for(i = 0; i < m->num; ++i) {
		if(m->count[i] == 0) {
			/* If this block was unused, we increase the reference count to
			   indicate that it now is used and return a pointer to the
			   memory block. */
			++(m->count[i]);
			return (void *)((char *)m->mem + (i * m->size));
		}
	}

#ifdef DEBUG_TELNETD
	uart_puts_P("telnetd: couldn't allocate line buffer");
#endif
	/* No free block was found, so we return NULL to indicate failure to
	   allocate block. */
	return NULL;
}
/*---------------------------------------------------------------------------*/
char memb_free(struct memb_blocks *m, void *ptr) {
	int i;
	char *ptr2;

	/* Walk through the list of blocks and try to find the block to
	   which the pointer "ptr" points to. */
	ptr2 = (char *)m->mem;
	for(i = 0; i < m->num; ++i) {
    
		if(ptr2 == (char *)ptr) {
			/* We've found to block to which "ptr" points so we decrease the
			   reference count and return the new value of it. */
			if(m->count[i] > 0) {
				/* Make sure that we don't deallocate free memory. */
				--(m->count[i]);
			}
			return m->count[i];
		}
		ptr2 += m->size;
	}
	return -1;
}
/*---------------------------------------------------------------------------*/
static char * alloc_line(void) {
	return memb_alloc(&linemem);
}
/*---------------------------------------------------------------------------*/
static void dealloc_line(char *line) {
	memb_free(&linemem, line);
}
/*---------------------------------------------------------------------------*/
static void help(struct telnetd_connection_state_t *state) {
#ifdef DEBUG_TELNETD
	uart_puts_P("telnetd: running command: help\r\n");
#endif
	shell_output(state, "Available commands:", "");
	shell_output(state, "io   -port io:", "");
	shell_output(state, "      io <port> <bit> <data>", "");
	shell_output(state, "rst  -reset rst <port> <bit>", "");
	shell_output(state, "help -show help", "");
	shell_output(state, "exit -exit shell", "");
	shell_output(state, "0 0:", telnetd_name[0]);
	shell_output(state, "0 1:", telnetd_name[1]);
	shell_output(state, "0 2:", telnetd_name[2]);
	shell_output(state, "0 3:", telnetd_name[3]);
	shell_output(state, "0 4:", telnetd_name[4]);
	shell_output(state, "0 5:", telnetd_name[5]);
	shell_output(state, "0 6:", telnetd_name[6]);
	shell_output(state, "0 7:", telnetd_name[7]);
}
/*---------------------------------------------------------------------------*/
static void unknown(struct telnetd_connection_state_t *state) {
#ifdef DEBUG_TELNETD
	uart_puts_P("telnetd: running command: unknown\r\n");
#endif
	if(state->bufptr > 0) {
		shell_output(state, "Unknown command: ", state->buf);
		help(state);
	}
}
/*---------------------------------------------------------------------------*/
char gethexnibble(uint8_t data) {
	if (data > 9) {
		data -= 10;
		return ('A' + data);
	} else {
		return ('0' + data);
	}
}
/*---------------------------------------------------------------------------*/
char * gethexbyte(uint8_t data) {
	/* first high, then low nibble */
	return (char *)(gethexnibble((data & 0xF0) >> 4)+gethexnibble(data & 0x0F));
}
/*---------------------------------------------------------------------------*/
static void telnetd_io(struct telnetd_connection_state_t *state) {
	u8_t data[3];
	u8_t j=0;
	char *parm;
	char *b=state->buf;

#ifdef DEBUG_TELNETD
	uart_puts_P("telnetd: running command: ");
	uart_puts(state->buf);
	uart_eol();
#endif

	//skip the telnet command itself
	while ((*b != '\0') && (*b != 0x20)) {
		b++;
	}

	if (*b != '\0') {
		parm=malloc(TELNETD_CONF_LINELEN);
		if (parm != NULL) {
			//get parameters
			parm[0]='\0';
			do {
				b++;
				if ((*b == 0x20) || (*b == 0x0)) { //space or end of line found: end of parameter
					int s = atoi(parm);
					if (((j == 0) && (s > 3)) || ((j == 1) && (s > 7)) || ((j == 2) && (s > 1))) {
#ifdef DEBUG_TELNETD
						uart_puts_P("telnetd: wrong value ");
						uart_puts(parm);
						uart_puts_P(" for parm ");
						uart_putdecbyte(j);
						uart_eol();
#endif
						j=255;
						break;
					} else {
						data[j]=s;
					}
					parm[0]='\0';
					j++;
				} else {
					int s = (int)(*b);
					if (s < 47 || s > 57) {
#ifdef DEBUG_TELNETD
						uart_puts_P("telnetd: wrong character ");
						uart_putc(*b);
						uart_puts_P(" for parm ");
						uart_putdecbyte(j);
						uart_eol();
#endif
						j=255;
						break;
					}
					char c[2];
					c[0]=*b;
					c[1]='\0';
					strcat(parm, c);
				}
			} while (*b != '\0');
			free(parm);
		}
	}

	if (j == 255) {
		shell_output(state, "error in command ", state->buf);

#ifdef USE_7SEGMENT
		strcpy(l7seg_buffer, "err   ");
#endif

	} else if (j<3) {
		shell_output(state, "too few parms", "");

#ifdef USE_7SEGMENT
		strcpy(l7seg_buffer, "par   ");
#endif

	} else {

#ifdef USE_7SEGMENT
		strcpy(l7seg_buffer, "io ");
		l7seg_buffer[3]=data[0];
		l7seg_buffer[4]=data[1];
		l7seg_buffer[5]=data[2];
#endif

	#ifdef SYSLOG_TELNETD
		syslog_message_telnet(0, data[0], data[1], data[2]);
	#endif

		//define the given bit as output
		uint8_t maddress=(data[0])*3+1;
		_SFR_IO8(maddress)|= (1<<data[1]);

		//set the bit to given value
		maddress=(data[0])*3+2;
		if (data[2]==0){  // reset bit
			_SFR_IO8(maddress)&= ~(1<<data[1]);
		} else {  // set bit
			_SFR_IO8(maddress)|= (1<<data[1]);
		}

	}

#ifdef USE_7SEGMENT
	l7dontdisplaytime=clock_global.current_time+10;
#endif
}
/*---------------------------------------------------------------------------*/
static void telnetd_rst(struct telnetd_connection_state_t *state) {
	u8_t data[2];
	u8_t j=0;
	char *parm;
	char *b=state->buf;

#ifdef DEBUG_TELNETD
	uart_puts_P("telnetd: running command: ");
	uart_puts(state->buf);
	uart_eol();
#endif

	//skip the telnet command rst itself
	while ((*b != '\0') && (*b != 0x20)) {
		b++;
	}

	if (*b != '\0') {
		parm=malloc(TELNETD_CONF_LINELEN);
		if (parm != NULL) {
			//get parameters
			parm[0]='\0';
			do {
				b++;
				if ((*b == 0x20) || (*b == 0x0)) { //space or end of line found: end of parameter
					int s = atoi(parm);
					if (((j == 0) && (s > 3)) || ((j == 1) && (s > 8))) {
#ifdef DEBUG_TELNETD
						uart_puts_P("telnetd: wrong value ");
						uart_puts(parm);
						uart_puts_P(" for parm ");
						uart_putdecbyte(j);
						uart_eol();
#endif
						j=255;
						break;
					} else {
						data[j]=s;
					}
					parm[0]='\0';
					j++;
				} else {
					int s = (int)(*b);
					if ((s < 47 || s > 57)) {
#ifdef DEBUG_TELNETD
						uart_puts_P("telnetd: wrong character ");
						uart_putc(*b);
						uart_puts_P(" for parm ");
						uart_putdecbyte(j);
						uart_eol();
#endif
						j=255;
						break;
					}
					char c[2];
					c[0]=*b;
					c[1]='\0';
					strcat(parm, c);
				}
			} while (*b != '\0');
			free(parm);
		}
	}

	if (j == 255) {
		shell_output(state, "error in command ", state->buf);

#ifdef USE_7SEGMENT
		strcpy(l7seg_buffer, "err   ");
#endif

	} else if (j<2) {
		shell_output(state, "too few parms", "");

#ifdef USE_7SEGMENT
		strcpy(l7seg_buffer, "par   ");
#endif

	} else {

#ifdef USE_7SEGMENT
		strcpy(l7seg_buffer, "res ");
		l7seg_buffer[4]=data[0];
		l7seg_buffer[5]=data[1];
#endif

#ifdef SYSLOG_TELNETD
		syslog_message_telnet(1, data[0], data[1], 0);
		syslog_message_telnetreset(telnetd_name[data[1]]);
#endif

		//define the given bit as output
		uint8_t maddress=(data[0])*3 + 0x1;
		_SFR_IO8(maddress)|= (1<<data[1]);

		//send a reset pulse to the given bit
		maddress=(data[0])*3+2;
		_SFR_IO8(maddress)|= (1<<data[1]);
		for(j=0;j<75;j++) {
			_delay_ms(10);
		}
		_SFR_IO8(maddress)&= ~(1<<data[1]);

		shell_output(state, "resetted host ", telnetd_name[data[1]]);
	}

#ifdef USE_7SEGMENT
	l7dontdisplaytime=clock_global.current_time+10;
#endif
}
/*---------------------------------------------------------------------------*/
void shell_quit(struct telnetd_connection_state_t *state) {
#ifdef DEBUG_TELNETD
	uart_puts_P("telnetd: running command: shell_quit\r\n");
#endif
	state->state = STATE_CLOSE;
}
/*---------------------------------------------------------------------------*/
static void parse(struct telnetd_connection_state_t *state, struct ptentry *t) {
	struct ptentry *p;
	char *cmd;
	u8_t i;

	cmd = malloc(TELNETD_CONF_LINELEN);
	if (cmd != NULL) {
		for (i=0;i<state->bufptr;i++) {
			cmd[i]=state->buf[i];
			if (state->buf[i] == 0x20) {
				break;
			}
		}
		cmd[i]='\0';
#ifdef DEBUG_TELNETD
		uart_puts_P("telnetd: got command '");
		uart_puts(state->buf);
		uart_puts_P("', using '");
		uart_puts(cmd);
		uart_puts_P("'");
		uart_eol();
#endif
		for(p = t; p->commandstr != NULL; ++p) {
			if(strncmp(p->commandstr, cmd, strlen(p->commandstr)) == 0) {
				break;
			}
		}
		free(cmd);

		p->pfunc(state);
#ifdef DEBUG_TELNETD
	} else {
		uart_puts_P("telnetd: couldn't allocate line buffer");
#endif
	}
}
/*---------------------------------------------------------------------------*/
void shell_input(struct telnetd_connection_state_t *state) {
	if (state->bufptr > 0) {
		parse(state, parsetab);
	}
	shell_prompt(state, SHELL_PROMPT);
}
/*---------------------------------------------------------------------------*/
static void get_char(struct telnetd_connection_state_t *state, u8_t c) {
	char *line;

//	if(c == ISO_cr) {
//		return;
//	}

	if(c == 0) {
		return;
	}

	//generate echo of input
	line = alloc_line();
	if(line != NULL) {
		if ((c == ISO_nl) || (c == ISO_cr)) {
			line[0] = ISO_cr;
			line[1] = ISO_nl;
			line[2] = 0;
		} else {
			line[0] = c;
			line[1] = 0;
		}
		sendline(state, line);
	}

	state->buf[(int)state->bufptr] = c;
	if(c == ISO_nl || c == ISO_cr || state->bufptr == sizeof(state->buf) - 1) {
		if(state->bufptr > 0) {
			state->buf[(int)state->bufptr] = 0;
		}
		shell_input(state);
		state->bufptr = 0;
	} else {
		++state->bufptr;
	}
}
/*---------------------------------------------------------------------------*/
static void sendline(struct telnetd_connection_state_t *state, char *line) {
	static unsigned int i;

	//search first empty line buffer
	for(i = 0; i < TELNETD_CONF_NUMLINES; ++i) {
		if(state->lines[i] == NULL) {
			//if found fill it with line contents
			state->lines[i] = line;
			break;
		}
	}
	if(i == TELNETD_CONF_NUMLINES) {
		dealloc_line(line);
	}
}
/*---------------------------------------------------------------------------*/
void shell_prompt(struct telnetd_connection_state_t *state, const char *str) {
	char *line;

	line = alloc_line();
	if(line != NULL) {
		strncpy(line, str, TELNETD_CONF_LINELEN);
		sendline(state, line);
	}
}
/*---------------------------------------------------------------------------*/
void shell_output(struct telnetd_connection_state_t *state, const char *str1, const char *str2) {
	static unsigned len;
	char *line;

	//allocate a line buffer
	line = alloc_line();
	if(line != NULL) {
		//copy contents of str1 (and, if it fits, also str2) into the line buffer
		len = strlen(str1);
		strncpy(line, str1, TELNETD_CONF_LINELEN);
		if(len < TELNETD_CONF_LINELEN) {
			strncpy(line + len, str2, TELNETD_CONF_LINELEN - len);
		}
		len = strlen(line);
		if(len < TELNETD_CONF_LINELEN - 2) {
			line[len] = ISO_cr;
			line[len+1] = ISO_nl;
			line[len+2] = 0;
		}
		//pass the line buffer to output queue
		sendline(state, line);
	}
}
/*---------------------------------------------------------------------------*/
void telnetd_init(void) {
	uip_listen(HTONS(TELNETD_PORT));
	memb_init(&linemem);
	eeprom_read_block(telnetd_name, (uint8_t *)EEPROM_CONFIG_NAMES, sizeof(struct eeprom_config_names_t));
}
/*---------------------------------------------------------------------------*/
static void sendopt(struct telnetd_connection_state_t *state, u8_t option, u8_t value) {
	char *line;

	line = alloc_line();
	if(line != NULL) {
		line[0] = TELNET_IAC;
		line[1] = option;
		line[2] = value;
		line[3] = 0;
		sendline(state, line);
	}
}
/*---------------------------------------------------------------------------*/
static void acked(struct telnetd_connection_state_t *state) {
	static unsigned int i;

#ifdef DEBUG_TELNETD
	u8_t j=0;
#endif
	//move all unsent lines forward (and deallocate line buffers of successfully sent lines)
	while(state->numsent > 0) {
		dealloc_line(state->lines[0]);
		for(i = 1; i < TELNETD_CONF_NUMLINES; ++i) {
			state->lines[i - 1] = state->lines[i];
		}
		state->lines[TELNETD_CONF_NUMLINES - 1] = NULL;
		--state->numsent;
#ifdef DEBUG_TELNETD
		j++;
#endif
	}
}
/*---------------------------------------------------------------------------*/
static void closed(struct telnetd_connection_state_t *state) {
	static unsigned int i;

	for(i = 0; i < TELNETD_CONF_NUMLINES; ++i) {
		if(state->lines[i] != NULL) {
			dealloc_line(state->lines[i]);
		}
	}
}
/*---------------------------------------------------------------------------*/
static void newdata(struct telnetd_connection_state_t *state) {
	u16_t len;
	u8_t c;
	char *dataptr;

	len = uip_datalen();
	dataptr = (char *)uip_appdata;
 
	while(len > 0 && state->bufptr < sizeof(state->buf)) {
		c = *dataptr;
		++dataptr;
		--len;
		switch(state->state) {
			case STATE_IAC:
				if(c == TELNET_IAC) {
					get_char(state, c);
					state->state = STATE_NORMAL;
				} else {
					switch(c) {
						case TELNET_WILL:
							state->state = STATE_WILL;
							break;
						case TELNET_WONT:
							state->state = STATE_WONT;
							break;
						case TELNET_DO:
							state->state = STATE_DO;
							break;
						case TELNET_DONT:
							state->state = STATE_DONT;
							break;
						default:
							state->state = STATE_NORMAL;
							break;
					}
				}
				break;
			case STATE_WILL:
				/* Reply with a DONT */
				sendopt(state, TELNET_DONT, c);
				state->state = STATE_NORMAL;
				break;
			case STATE_WONT:
				/* Reply with a DONT */
				sendopt(state, TELNET_DONT, c);
				state->state = STATE_NORMAL;
				break;
			case STATE_DO:
				/* Reply with a WONT */
				sendopt(state, TELNET_WONT, c);
				state->state = STATE_NORMAL;
				break;
			case STATE_DONT:
				/* Reply with a WONT */
				sendopt(state, TELNET_WONT, c);
				state->state = STATE_NORMAL;
				break;
			case STATE_NORMAL:
				if(c == TELNET_IAC) {
					state->state = STATE_IAC;
				} else {
					get_char(state, c);
				}
				break;
		}
	}
}
/*---------------------------------------------------------------------------*/
static void senddata(struct telnetd_connection_state_t *state) {
	static char *bufptr, *lineptr;
	static int buflen, linelen;

	bufptr = uip_appdata;
	buflen = 0;
	//check all filled line buffers
	for(state->numsent = 0; state->numsent < TELNETD_CONF_NUMLINES && 	state->lines[state->numsent] != NULL ; ++state->numsent) {
		lineptr = state->lines[state->numsent];
		linelen = strlen(lineptr);
		if(linelen > TELNETD_CONF_LINELEN) {
			linelen = TELNETD_CONF_LINELEN;
		}
		if(buflen + linelen < uip_mss()) {
			memcpy(bufptr, lineptr, linelen);
			bufptr += linelen;
			buflen += linelen;
		} else {
			break;
		}
	}
	if (buflen != 0) {
		uip_send(uip_appdata, buflen);
	}
}
/*---------------------------------------------------------------------------*/
void telnetd_main(void) {
	static unsigned int i;

	struct telnetd_connection_state_t *state = &uip_conn->appstate.telnetd;

	if(uip_connected()) {
#ifdef DEBUG_TELNETD
		uart_puts_P("telnetd: new connection\r\n");
#endif
#ifdef SYSLOG_TELNETD
		uip_ipaddr_t *remoteip = &uip_conn->ripaddr;
	    uint8_t *byte = (uint8_t *)remoteip;
		syslog_message_telnetnew(byte[0], byte[1], byte[2], byte[3]);
#endif
		for(i = 0; i < TELNETD_CONF_NUMLINES; ++i) {
			state->lines[i] = NULL;
		}
		state->bufptr = 0;
		state->state = STATE_NORMAL;

		//shell_start
		shell_output(state,"uIP command shell", "");
		shell_output(state,"Type '?' and return for help", "");
		shell_prompt(state,SHELL_PROMPT);
	}

	if(state->state == STATE_CLOSE) {
#ifdef DEBUG_TELNETD
		uart_puts_P("telnetd: closing connection\r\n");
#endif
#ifdef SYSLOG_TELNETD
		syslog_message_P("telnetd: closing connection");
#endif
		state->state = STATE_NORMAL;
		uip_close();
		return;
	}

	if(uip_closed() || uip_aborted() || uip_timedout()) { 
#ifdef DEBUG_TELNETD
		uart_puts_P("telnetd: closed\r\n");
#endif
#ifdef SYSLOG_TELNETD
		syslog_message_P("telnetd: closed");
#endif
		closed(state);
	}

	if(uip_acked()) {
#ifdef DEBUG_TELNETD
		uart_puts_P("telnetd: acked\r\n");
#endif
		acked(state);
	}

	if(uip_newdata()) {
#ifdef DEBUG_TELNETD
		uart_puts_P("telnetd: got new data\r\n");
#endif
		newdata(state);
	}

	if(uip_rexmit() || uip_newdata() || uip_acked() || uip_connected() || uip_poll()) {
#ifdef DEBUG_TELNETD
//		uart_puts_P("telnetd: sending data\r\n");
#endif
		senddata(state);
	}
}
/*---------------------------------------------------------------------------*/
